#!/usr/local/gauche/bin/gosh
;; -*- coding: utf-8-unix -*-

(add-load-path ".." :relative #t)
(use tar)
(use gauche.uvector)
(use gauche.collection)

;;;
;;; Extract file from tar archive with generator method.
;;; Extract with generation can only work with regular file not include
;;; GNU sparse file.
;;; If you use it cooperates with random access, multiple generator can be
;;; gotten at the same time.
;;; 
(define *entry-gen1* #f)
(define *entry-gen2* #f)
(let* ((tar (open-input-tar "../testdata/gpl.tar" :random-access #t))
       (tar-next (tar->generator tar)))
  (set! *entry-gen1* (extract-generator (tar-next)))
  (set! *entry-gen2* (extract-generator (tar-next))))

(generator-for-each (^x (format #t "~a" (u8vector->string x))) *entry-gen1*)
(generator-for-each (^x (format #t "~a" (u8vector->string x))) *entry-gen2*)
(format #t "~a" (let ((v (*entry-gen1*)))
                  (if (eof-object? v) "end of object." v)))


;;;
;;; Extract file from tar archive with generator method.
;;; Following code gets generator sequentially that order is determined by the
;;; entry order in the tar archive.
;;; 
(guard
    (exc
     ((tar-error? exc)
      (format (current-error-port) "~a (~a)\n"
              (tar-error-message exc) (tar-error-debug-info exc))))
  (let* ((tar (open-input-tar "../testdata/gpl.tar"))
         (tar-next (tar->generator tar))
         (gen #f)
         (entry #f))
    (set! entry (tar-next))
    (set! gen (extract-generator entry))
    (generator-for-each (^x (format #t "~a" (u8vector->string x))) gen)
    (set! entry (tar-next))
    (set! gen (extract-generator entry))
    (generator-for-each (^x (format #t "~a" (u8vector->string x))) gen)))

;;;
;;; Extract file from tar+gz archive with generator method.
;;; When you will get generator, you can set buffer size which uses when reading
;;; tar entry for each generator calls.
;;; Buffer size must be multiple of 512 byte because tar file format oblige to
;;; us the tar entry size is 512 or multiple of 512.
;;; 
(guard
    (exc
     ((tar-error? exc)
      (format (current-error-port) "~a (~a)\n"
              (tar-error-message exc) (tar-error-debug-info exc))))
  (let* ((tar (open-input-tar "../testdata/gpl.tgz"))
         (tar-next (tar->generator tar))
         (gen #f)
         (entry #f))
    (set! entry (tar-next))
    (set! gen (extract-generator entry :buffer-size (* 512 2)))
    (generator-for-each (^x (format #t "~a" (u8vector->string x))) gen)
    (set! entry (tar-next))
    ;;; 
    ;;; Buffer size can't accept negative value.
    ;;; 
    ;; (set! gen (extract-generator entry :buffer-size (* 512 2 -1)))

    ;;; 
    ;;; Buffer size which exceeds far actual tar entry size can accept and
    ;;; is adjusted internal of function. As same as it, 0 can accept too.
    ;;; 
    ;; (set! gen (extract-generator entry :buffer-size (* 512 2000)))
    ;; (set! gen (extract-generator entry :buffer-size 0))

    (set! gen (extract-generator entry :buffer-size (* 512 2)))
    (generator-for-each (^x (format #t "~a" (u8vector->string x))) gen)))


