#!/usr/local/gauche/bin/gosh
;; -*- coding: utf-8-unix -*-

(add-load-path ".." :relative #t)
(use tar)
(use gauche.collection)

;;; 
;;; Tar archive iteration.
;;; 
(let ((tar (open-input-tar "./ztestdataposix.tar")))
  (unwind-protect
   (map (cut format #t "~S" <>) tar)
   (close-input-tar tar)))

;;; 
;;; Tar archive iteration with call-with-*. Same as above code.
;;; 
(call-with-input-tar "./ztestdataposix.tar"
                     (^(tar) (map (cut format #t "~S" <>) tar)))

;;; 
;;; Debug tar reading. Show tar raw header only.
;;; 
(let ((tar (open-input-tar "../testdata/ztestdataposix.tar" :debug #t)))
  (unwind-protect
   (map describe tar)
   (close-input-tar tar)))

;;; 
;;; Tar iteration using generator.
;;; 
(let* ((tar (open-input-tar "../testdata/ztestdataposix.tar"))
       (tar-next (tar->generator tar)))
  (describe tar)
  (unwind-protect
   (begin
     (format #t "~S" (tar-next))
     (format #t "~S" (tar-next)))
   (close-input-tar tar)))

(let* ((tar (open-input-tar "../testdata/ztestdataposix.tar"))
       (tar-next (tar->generator tar)))
  (unwind-protect
   (generator-for-each (^x (format #t "~a\n" (~ x 'name))) tar-next)
   (format #t "is closed? ~S\n" (~ tar 'is-closed?))
   (format #t "is generate? ~S\n" (~ tar 'in-generation?))
   (close-input-tar tar)))

;;; 
;;; Debug tar reading using by generator.
;;; 
(let* ((tar (open-input-tar "../testdata/ztestdataposix.tar" :debug #t))
       (tar-next (tar->generator tar)))
  (unwind-protect
   (begin
    (describe (tar-next))
    (describe (tar-next)))
   (close-input-tar tar)))

;;; 
;;; Generator restart.
;;; 
(let* ((tar (open-input-tar "../testdata/ztestdataposix.tar"))
       (tar-next (tar->generator tar)))
  (unwind-protect
   (begin
     (format #t "~S" (tar-next))
     (format #t "~S" (tar-next))
     (if (~ tar 'in-generation?)
       (format #t "\n**tar ~S is in the generation.**\n" tar))
     (set! tar-next (tar->generator tar))
     (format #t "~S" (tar-next))
     (format #t "~S" (tar-next)))
   (close-input-tar tar)))

;;; 
;;; Extract data from tar to disk.
;;; 
(let* ((tar (open-input-tar "../testdata/ztestdataposix.tar"))
       (tar-next (tar->generator tar)))
  (unwind-protect
   (generator-for-each extract tar-next)
  (close-input-tar tar)))

(let ((tar (open-input-tar "../testdata/zlonggnu.tar")))
  (unwind-protect
   (map (^x
         (format #t "~S" x)
         (extract x)) tar)
   (close-input-tar tar)))

;;; 
;;; Extract data from tar to buffer.
;;; 
(let* ((tar (open-input-tar "../testdata/ztestdataposix.tar"))
       (tar-next (tar->generator tar))
       (out (open-output-string)))
  (unwind-protect
   (generator-for-each
    (^x
     (if (#/archive\.c/ (~ x 'name))
       (extract x out)))
    tar-next)
   (format #t "output to buffer [~a]\n" (get-output-string out))
  (close-input-tar tar)))

;;; 
;;; Read tar-gzip.
;;; 
(let ((tar-gz (open-input-tar-gz "posix-build.tar.gz")))
  (guard
      (exc
       ((tar-error? exc)
        (format #t "~a ~a\n"
                (tar-error-message exc)
                (tar-error-debug-info exc))))
    (map (^x (format #t "~S" x)) tar-gz)))

;;; 
;;; Debug tar-gzip.
;;; 
(let ((tar-gz (open-input-tar-gz "posix-build.tar.gz" :debug #t)))
  (guard
      (exc
       ((tar-error? exc)
        (format #t "~a ~a\n"
                (tar-error-message exc)
                (tar-error-debug-info exc))))
    (map describe tar-gz)))

;;; 
;;; Tar-gzip iteration using generator.
;;; 
(let* ((tar-gz (open-input-tar-gz "posix-build.tar.gz"))
       (tar-gz-next (tar->generator tar-gz)))
  (guard
      (exc
       ((tar-error? exc)
        (format #t "~a ~a\n"
                (tar-error-message exc)
                (tar-error-debug-info exc))
        (close-input-tar-gz tar-gz)))
    (format #t "~S" (tar-gz-next))
    (format #t "~S" (tar-gz-next))

    ;;; gzip doesn't support generator restart.
    ;;; Raise <tar-error>.
    (set! tar-gz-next (tar->generator tar-gz)) 
    (format #t "~S" (tar-gz-next))))

(let* ((tar-gz (open-input-tar-gz "posix-build.tar.gz"))
       (tar-gz-next (tar->generator tar-gz)))
  (unwind-protect
   (guard
      (exc
       ((tar-error? exc)
        (format #t "~a ~a\n"
                (tar-error-message exc)
                (tar-error-debug-info exc))))
     (generator-for-each (cut format #t "~S" <>) tar-gz-next))
   (close-input-tar-gz tar-gz)))


;;; 
;;; Tar with random access mode. This mode can read data of where any position
;;; of Tar.
;;; Warning: This mode is under experimental.
;;; 
(let* ((tar (open-input-tar "../testdata/ztestdataposix.tar" :random-access #t))
       (tar-next (tar->generator tar))
       (out (open-output-string))
       (entry #f))
  (unwind-protect
   (generator-for-each
    (^x (if (#/archive_class\.c/ (~ x 'name)) (set! entry x)))
    tar-next)
   (format #t "~S" entry)
   (extract entry out)
   (format #t "output to buffer [~a]\n" (get-output-string out))
  (close-input-tar tar)))


