#!/usr/local/gauche/bin/gosh
;; -*- coding: utf-8-unix -*-

(add-load-path ".." :relative #t)
(use tar)
(use file.util)
(use gauche.collection)

;;;
;;; Create tar. <tar-writer> writes content at once.
;;; 
(let ((tar-writer (open-output-tar-writer "posix-write.tar")))
  (unwind-protect
   (begin
     (add-content tar-writer "tarread.scm")
     (add-content tar-writer "tarwrite.scm")))
  (close-output-tar-writer tar-writer))

;;;
;;; Create tar. <tar-builder> writes content each calls of `add-content`.
;;; 
(let ((tar-builder (open-output-tar-builder "posix-build.tar")))
  (define (list-files root matcher proc)
    (directory-fold root (^(src seed) (if (matcher src) (proc src) seed)) '()))

  (unwind-protect
   (list-files "." #/.*\.c/ (^(src) (add-content tar-builder src)))
   (close-output-tar-builder tar-builder)))

;;;
;;; Create tar-gzip. <tar-gz-writer> is gzip version of <tar-writer>.
;;; 
(let ((tar-gz-writer (open-output-tar-gz-writer "posix-write.tar.gz"
                                                :compression-level 9)))
  (unwind-protect
   (begin
     (add-content tar-gz-writer "tarread.scm")
     (add-content tar-gz-writer "tarwrite.scm")))
  (close-output-tar-gz-writer tar-gz-writer))

;;;
;;; Create tar-gzip. <tar-gz-builder> is gzip version of <tar-builder>.
;;; 
(let ((tar-gz-builder (open-output-tar-gz-builder "posix-build.tar.gz")))
  (define (list-files root matcher proc)
    (directory-fold root (^(src seed) (if (matcher src) (proc src) seed)) '()))

  (unwind-protect
   (list-files "." #/.*\.c/ (^(src) (add-content tar-gz-builder src)))
   (close-output-tar-gz-builder tar-gz-builder)))
