#!/usr/local/gauche/bin/gosh
;; -*- coding: utf-8-unix -*-

;;; one line to give the Gauche-tar and an idea of what it does.
;;; Copyright (C) 2018 aaauuue 
;;;
;;; This program is free software; you can redistribute it and/or
;;; modify it under the terms of the GNU General Public License
;;; as published by the Free Software Foundation; either version 2
;;; of the License, or (at your option) any later version.
;;;
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with this program; if not, write to the Free Software
;;; Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

(select-module tar)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Read tar
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(define (open-input-tar src :key (debug #f) (random-access #f) (fast-skip? #f))
  (let ((tar #f))
    (cond
     ((and (eq? debug #t) (eq? random-access #f))
      (set! tar (make <debug-tar-reader>)))
     ((eq? random-access #t)
      (set! tar (make <tar-random-reader>)))
     (else (set! tar (make <tar-reader>))))
    
    (set! (~ tar 'name) src)
    (set! (~ tar 'ioport) (open-input-file src))
    (set! (~ tar 'fast-skip?) fast-skip?)
    tar))

(define (open-input-tar-from-port port :key (debug #f) (random-access #f) (fast-skip? #f))
  (let ((tar #f))
    (if (not (eq? (input-port? port) #t))
      (tar-condition-throw
       (<tar-error>
        "argument `port` must be input-port."
        "open-input-tar-from-port port :key (debug #f) (random-access #f)")))

    (cond
     ((and (eq? debug #t) (eq? random-access #f))
      (set! tar (make <debug-tar-reader>)))
     ((eq? random-access #t)
      (set! tar (make <tar-random-reader>)))
     (else (set! tar (make <tar-reader>))))
    
    (set! (~ tar 'name) (port-name port))
    (set! (~ tar 'ioport) port)
    (set! (~ tar 'fast-skip?) fast-skip?)
    tar))

(define-method close-input-tar ((tar <tar-reader-base>))
  (close-input-port (~ tar 'ioport))
  (set! (~ tar 'name) "")
  (if (not (eq? (~ tar 'post-processes) '()))
    (dolist (f (~ tar 'post-processes))
      (apply (car f) (cdr f)))))

(define-macro (tar-debug-generator end-func)
  `(lambda ()
     (if (and (eq? (~ tar 'skip-data?) #t)
              (not (eq? tar-header #f)))
       (begin
         (skip-tar-data tar tar-header)
         (set! (~ tar 'skip-data?) #t)))
     (set! tar-header (read-tar-header tar))
     (,end-func tar tar-header)))

(define-method call-with-iterator ((tar <debug-tar-reader-base>) proc . opts)
  (let ((tar-header #f)
        (continue-sparse #f))
    (proc 
     (tar-debug-generator end-of-tar?)
     (lambda () tar-header))))

(define-method tar->generator ((tar <debug-tar-reader-base>))
  (let ((tar-header #f)
        (continue-sparse #f))
    (if (and (is-a? tar <tar-gz-base>) (eq? (~ tar 'in-generation?) #t))
      (tar-condition-throw
       (<tar-error> "tar with gzip doesn't reset generator condition."
                    "tar->generator ((tar <debug-tar-reader-base>))")))
    (port-seek (~ tar 'ioport) 0 SEEK_SET)
    (set! (~ tar 'in-generation?) #t)
    (tar-debug-generator
     (^(tar tar-header)
       (if (end-of-tar? tar tar-header)
         (begin
           (set! (~ tar 'in-generation?) #f)
           (eof-object))
         tar-header)))))

(define-macro (tar-generator end-func)
  `(lambda ()
       (define (build-sparse-info pos-list realsize)
         (let ((last-pos 0)
               (result '()))
           (dolist (pos pos-list)
             (match pos
               ((offset numbytes)
                (set! result (append (list (list last-pos offset
                                                 (- offset last-pos) 'hole))
                                     result))
                (set! result (append (list (list offset (+ offset numbytes)
                                                 (- (+ offset numbytes) offset)
                                                 'fill))
                                     result))
                (set! last-pos (+ offset numbytes)))))
           (if (< last-pos realsize)
             (set! result (append (list (list last-pos) realsize
                                        (- realsize last-pos) 'hole))))
           (reverse result)))

       (define (read-more-tar-header)
         (if (and (eq? (~ tar 'skip-data?) #t)
                  (not (eq? tar-header #f)))
           (begin
             (skip-tar-data tar tar-header)
             (set! (~ tar 'skip-data?) #t)))
         (set! tar-header (read-tar-header tar)))

       (define (copy-pax-entry entry-alist dest)
         (dolist (item entry-alist)
           (match item
             ((num key value)
              (cond
               ((#/^GNU.*/ (symbol->string key))
                (cond
                 ((or (#/GNU\.sparse\.offset/ (symbol->string key))
                      (#/GNU\.sparse\.numbytes/ (symbol->string key)))
                  (set! (~ dest 'posix-gnu-keywords key)
                        (append (list value) (~ dest 'posix-gnu-keywords key))))
                 (else 
                  (if (class-slot-definition
                       (class-of #0=(~ dest 'posix-gnu-keywords)) key)
                    (set! (~ #0# key) value)))))
               ((or (#/^RHT\.security\.selinux/ (symbol->string key))
                    (#/^SCHILY\.acl\..*/ (symbol->string key)))
                (if (class-slot-definition
                     (class-of #1=(~ dest 'posix-extra-keywords)) key)
                  (set! (~ #1# key) value)))
               ((#/SCHILY.xattr\..*/ (symbol->string key))
                (let ((xattr-list (~ tar-header 'posix-xattr-list)))
                  (set xattr-list (append (list (cons key value)) xattr-list))))
               (else
                (if (class-slot-definition
                     (class-of #2=(~ dest 'posix-default-keywords)) key)
                  (set! (~ #2# key) value))))))))

       (define (merge-entries src dest)
         (dolist (slot (map slot-definition-name (class-slots (class-of src))))
           (if (class-slot-definition (class-of dest) slot)
             (cond
              ((member slot '(atime ctime mtime))
               (if (not (eq? (~ src slot) #f))
                 (set! (~ dest slot) (car (~ src slot)))))
              (else
               (if (not (eq? (~ src slot) #f))
                 (set! (~ dest slot) (~ src slot)))))
             (begin
               (if (and (eq? slot 'linkpath)
                        (not (eq? (~ src 'linkpath) #f)))
                 (set! (~ dest 'linkname) (~ src 'linkpath)))
               (if (and (eq? slot 'path)
                        (not (eq? (~ src 'path) #f)))
                 (set! (~ dest 'name) (~ src 'path)))))))

       (define (dispatch-header)
         (define (more-old-gnu-sparse)
           (let ((sparse (make <gnu-old-sparse-header>))
                 (offset 0)
                 (numbytes 0)
                 (pos-list '()))
             (uvector-copy! (~ sparse 'header) 0
                            (read-uvector <u8vector> *tar-block-size*
                                          (~ tar 'ioport)))
             (call/cc
              (^(out)
                (dolist (s (~ sparse 'sp))
                  (set! offset (~ s 'offset))
                  (set! numbytes (~ s 'numbytes))
                  (if (and (= offset 0) (= numbytes 0)) (out #f))
                  (set! pos-list
                        (append (list (list offset numbytes)) pos-list)))
                (if (= (~ sparse 'isextended) 1)
                  (set! pos-list
                        (concatenate (list pos-list (more-old-gnu-sparse)))))))
             (reverse pos-list)))
         (cond
          ((string=? (~ tar-header 'typeflag) "L")
           (set! (~ tar-entry 'name) (get-tar-long-name tar tar-header))
           (read-more-tar-header)
           (dispatch-header))
          ((string=? (~ tar-header 'typeflag) "K")
           (set! (~ tar-entry 'linkname) (get-tar-long-name tar tar-header))
           (read-more-tar-header)
           (dispatch-header))
          ((string=? (~ tar-header 'typeflag) "S")
           (let ((tar-gheader (make <tar-header-gnuold>))
                 (offset 0)
                 (numbytes 0)
                 (pos-list '()))
             (uvector-copy! (~ tar-gheader 'header) 0 (~ tar-header 'header))
             (call/cc
              (^(out)
                (dolist (s (~ tar-gheader 'sparse))
                  (set! offset (~ s 'offset))
                  (set! numbytes (~ s 'numbytes))
                  (if (and (= offset 0) (= numbytes 0)) (out #f))
                  (set! pos-list
                        (append pos-list (list (list offset numbytes)))))))
             (if (= (~ tar-gheader 'isextended) 1)
               (set! pos-list
                     (concatenate (list pos-list (more-old-gnu-sparse)))))
             (set! tar-header tar-gheader)
             (set! (~ tar-entry 'sparse-info)
                   (build-sparse-info pos-list (~ tar-header 'realsize)))))
          ((string=? (~ tar-header 'typeflag) "x") ;;; posix extend header
           (let ((out (open-output-string)))
             (extract-tar-data-to-buffer tar tar-header out)
             (copy-pax-entry (parse-pax-header (get-output-string out)) tar-entry)
             ;;
             ;; GNU sparse information format version 0.0
             ;;
             (if (not (eq? (~ tar-entry 'posix-gnu-keywords 'GNU.sparse.map) #f))
               (set! (~ tar-entry 'sparse-info)
                     (build-sparse-info
                      (parse-gnu-sparse-map
                       (~ tar-entry 'posix-gnu-keywords 'GNU.sparse.map))
                      (~ tar-entry 'posix-gnu-keywords 'GNU.sparse.size))))
             ;;
             ;; GNU sparse information format version 0.1
             ;;
             (if (and
                  (not (eq?
                        (~ tar-entry 'posix-gnu-keywords 'GNU.sparse.offset)
                        '()))
                  (not (eq?
                        (~ tar-entry 'posix-gnu-keywords 'GNU.sparse.numbytes)
                        '())))
               (set! (~ tar-entry 'sparse-info)
                     (build-sparse-info
                      (zip
                       (reverse (~ tar-entry 'posix-gnu-keywords 'GNU.sparse.offset))
                       (reverse (~ tar-entry 'posix-gnu-keywords 'GNU.sparse.numbytes)))
                      (~ tar-entry 'posix-gnu-keywords 'GNU.sparse.size))))
             (read-more-tar-header)
             ;;
             ;; GNU sparse information format version 1.0
             ;;
             (if (not (eq?
                       (~ tar-entry 'posix-gnu-keywords 'GNU.sparse.major)
                       #f))
               (let ((lst '()) (x "") (y "") (count "") (byte-count 0))
                 (set! count (read-line (~ tar 'ioport)))
                 (dotimes (i (x->integer count))
                   (set! x (read-line (~ tar 'ioport)))
                   (set! y (read-line (~ tar 'ioport)))
                   (set! lst
                         (append (list (list (x->integer x) (x->integer y)))
                                 lst))
                   (set! byte-count
                         (+ byte-count (string-length x) (string-length y) 2)))
                 (set! byte-count (+ byte-count 1 (string-length count)))
                 (set! (~ tar-entry 'sparse-info)
                       (build-sparse-info
                        (reverse lst)
                        (~ tar-entry 'posix-gnu-keywords 'GNU.sparse.realsize)))
                 (port-seek (~ tar 'ioport) (* -1 byte-count) SEEK_CUR)
                 (set! (~ tar-entry 'sparse-byte-offset) byte-count)))
             (dispatch-header)))
          ((string=? (~ tar-header 'typeflag) "g") ;;; posix global extend header
           (let ((out (open-output-string)))
             (set! global-tar-entry (make <tar-entry>))
             (extract-tar-data-to-buffer tar tar-header out)
             (copy-pax-entry (parse-pax-header (get-output-string out))
                             global-tar-entry)
             (read-more-tar-header)
             (dispatch-header)))))
       
       (set! tar-entry (make <tar-entry>))
       (set! (~ tar-entry 'tarobj) tar)
       (read-more-tar-header)
       (if (and (not (= (~ tar-header 'chksum) 0))
                (not (= (calc-chksum tar-header) (~ tar-header 'chksum))))
         (tar-condition-throw
          (<tar-error>
           "header check sum has been discrepancy."
           #"original: ~(~ tar-header 'chksum) calc: ~(calc-chksum tar-header)")))

       (dispatch-header)
       (copy-class-slot-value tar-header tar-entry)
       (merge-entries (~ tar-entry 'posix-default-keywords) tar-entry)

       ;;; 
       ;;; For posix format longname.
       ;;; This code is under testing because we can't make tar archive
       ;;; with long name of it.
       ;;; 
       ;; (if (and #?,(eq? (tar-type tar-header) 'tar-posix)
       ;;          (not (= (~ tar-entry 'tarobj 'prefix) 0)))
       ;;   (set! (~ tar-entry name)
       ;;         (format #f "~a/~a" (~ tar-header 'prefix) (~ tar-entry 'name))))

       ;;; copy sparse realsize and name
       (if (not (eq? (~ tar-entry 'posix-gnu-keywords 'GNU.sparse.size) #f))
         (set! (~ tar-entry 'realsize)
               (~ tar-entry 'posix-gnu-keywords 'GNU.sparse.size)))
       (if (not (eq? (~ tar-entry 'posix-gnu-keywords 'GNU.sparse.realsize) #f))
         (set! (~ tar-entry 'realsize)
               (~ tar-entry 'posix-gnu-keywords 'GNU.sparse.realsize)))
       (if (not (eq? (~ tar-entry 'posix-gnu-keywords 'GNU.sparse.name) #f))
         (set! (~ tar-entry 'name)
               (~ tar-entry 'posix-gnu-keywords 'GNU.sparse.name)))
       (if (not (eq? global-tar-entry #f))
         (merge-entries (~ tar-entry 'posix-default-keywords) global-tar-entry))
       (set! (~ tar-entry 'random-access-offset) (port-tell (~ tar 'ioport)))
       (,end-func tar tar-header)))

(define-method call-with-iterator ((tar <tar-reader-base>) proc . opts)
  (let ((tar-header #f)
        (tar-entry #f)
        (global-tar-entry #f))
    (proc 
     (tar-generator end-of-tar?)
     (lambda () tar-entry))))

(define-method tar->generator ((tar <tar-reader-base>))
  (let ((tar-header #f)
        (tar-entry #f)
        (global-tar-entry #f))
    (if (and (is-a? tar <tar-gz-base>) (eq? (~ tar 'in-generation?) #t))
      (tar-condition-throw
       (<tar-error> "tar with gzip doesn't reset generator condition."
                    "tar->generator ((tar <tar-reader-base>))")))
    (port-seek (~ tar 'ioport) 0 SEEK_SET)
    (set! (~ tar 'in-generation?) #t)
    (tar-generator
     (^(tar tar-header)
       (if (end-of-tar? tar tar-header)
         (begin
           (set! (~ tar 'in-generation?) #f)
           (eof-object))
         tar-entry)))))

(define (call-with-input-tar src proc :key (debug #f) (random-access #f) (fast-skip? #f))
  (let ((tar (open-input-tar
              src :debug debug :random-access random-access
              :fast-skip? fast-skip?)))
    (unwind-protect
     (proc tar)
     (close-input-tar tar))))

(define (call-with-input-tar-from-port port proc :key (debug #f) (random-access #f) (fast-skip? #f))
  (let ((tar (open-input-tar-from-port
              port :debug debug :random-access random-access
              :fast-skip? fast-skip?)))
    (unwind-protect
     (proc tar)
     (close-input-tar tar))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Extract
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(define-method make-tar-directory ((tar <tar>) (entry <tar-entry>))
  (make-directory* (~ entry 'name))
  (set! (~ tar 'skip-data?) #t))

(define-method extract-tar-regular-file ((tar <tar>) (entry <tar-entry>) :key (output #f))
  (let* ((name (~ entry 'name))
         (block-count (tar-block-count #0=(~ entry 'size)))
         (remain 0)
         (count 0)
         (buf #f)
         (out (or output (open-output-file name))))
    ;; (format (current-error-port) "~S\n" (~ entry 'sparse-info))
    (if (and (not (= #0# 0)) (eq? (~ entry 'sparse-info) '()))
      (begin
        (set! remain (- 512 (- (* block-count *tar-block-size*) #0#)))
        (unwind-protect
         (begin 
           (dotimes (i (- block-count 1))
             (begin
               (set! buf (read-uvector <u8vector> *tar-block-size*
                                       (~ tar 'ioport)))
               (write-uvector buf out 0 *tar-block-size*)
               (uvector-copy! buf 0 *tar-eob*)))
           (begin
             (if (not (eq? buf #f)) (uvector-copy! buf 0 *tar-eob*))
             (set! buf (read-uvector <u8vector> *tar-block-size*
                                     (~ tar 'ioport)))
             (write-uvector buf out 0 remain)))
         (if (eq? output #f) (close-output-port out))))
      (begin
        (port-seek (~ tar 'ioport)
                   (* (tar-block-count (~ entry 'sparse-byte-offset))
                      *tar-block-size*)
                   SEEK_CUR)
        (dolist (sp (~ entry 'sparse-info))
          (match sp
            ((beg end size state)
             (cond
              ((eq? state 'fill)
               (if (not (= size 0))
                 (begin
                   (set! buf (make-uvector <u8vector> size 0))
                   (set! buf (read-uvector <u8vector> size (~ tar 'ioport)))
                   (write-uvector buf out 0 size))
                 ;; (set! buf (make-uvector <u8vector> *tar-block-size* 0))
                 ;; (begin
                 ;;   (set! count (div size *tar-block-size*))
                 ;;   (set! remain (mod size *tar-block-size*))
                 ;;   (dotimes (i count)
                 ;;     (set! buf (read-uvector <u8vector> *tar-block-size* (~ tar 'ioport)))
                 ;;     (write-uvector buf out 0 *tar-block-size*)
                 ;;     (uvector-copy! buf 0 *tar-eob*))
                 ;;   (begin
                 ;;     (if (not (eq? buf #f)) (uvector-copy! buf 0 *tar-eob*))
                 ;;     (set! buf (read-uvector <u8vector> *tar-block-size* (~ tar 'ioport)))
                 ;;     (write-uvector buf out 0 remain)))
                 ))
              ((eq? state 'hole)
               (if (not (= size 0))
                 (begin
                   (set! buf (make-uvector <u8vector> size 0))
                   (write-uvector buf out 0 size))
                 ;; (set! buf (make-uvector <u8vector> *tar-block-size* 0))
                 ;; (begin
                 ;;   (set! count (div size *tar-block-size*))
                 ;;   (set! remain (mod size *tar-block-size*))
                 ;;   (begin
                 ;;     (uvector-copy! buf 0 *tar-eob*)
                 ;;     (dotimes (i count)
                 ;;       (write-uvector buf out 0 *tar-block-size*)
                 ;;       (uvector-copy! buf 0 *tar-eob*))
                 ;;     (begin
                 ;;       (uvector-copy! buf 0 *tar-eob*)
                 ;;       (write-uvector buf out 0 remain))))
                 ))))))
        (if (eq? output #f) (close-output-port out))
        (let ((block-actual-size (* *tar-block-size*
                                    (tar-block-count (~ entry 'realsize)))))
          (if (not (= (~ entry 'realsize) block-actual-size))
          (port-seek (~ tar 'ioport)
                     (- block-actual-size (~ entry 'realsize))
                     SEEK_CUR)))))
    (set! (~ tar 'skip-data?) #f)))

(define-method extract-tar-regular-file-generator ((tar <tar>) (entry <tar-entry>) buffer-size)
  (letrec ((iter
            (lambda (return)
              (let loop ((buf-size 0)
                         (block-count 0)
                         (remain 0)
                         (v #f)
                         (block-counter 0))
                (set! buf-size buffer-size)
                (set! block-count (div (+ (~ entry 'size) (- buffer-size 1))
                                       buffer-size))
                (set! remain (- buffer-size
                                (- (* block-count buffer-size) (~ entry 'size))))
                (if (= block-counter 0)
                  (if (is-a? (~ entry 'tarobj) <tar-random-reader-base>)
                    (port-seek (~ entry 'tarobj 'ioport)
                               (~ entry 'random-access-offset)
                               SEEK_SET)))
                (if (not (= block-count block-counter))
                  (if (not (= (- block-count 1) block-counter))
                    (set! v (read-uvector <u8vector> buf-size (~ tar 'ioport)))
                    (let ((tmp-v (read-uvector <u8vector> buf-size (~ tar 'ioport))))
                      (set! v (u8vector-copy tmp-v 0 remain)))))
                (if (and (not (eof-object? v))
                         (not (= block-count block-counter)))
                  (begin
                    (set! return
                          (call/cc
                           (lambda (cont)
                             (set! iter cont)
                             (set! block-counter (+ block-counter 1))
                             (return v))))
                    (loop buf-size block-count remain v block-counter))
                  (begin
                    (set! (~ tar 'skip-data?) #f)
                    (return (eof-object)))))
              (begin
                (set! (~ tar 'skip-data?) #f)
                (return (eof-object))))))
    (lambda ()
      (call/cc
       (lambda (cont) (iter cont))))))


(define (extract-tar-symboliclink name linkname) 
  (define-syntax with-current-directory
    (syntax-rules ()
      ((_ body ...)
       (let ((cur (current-directory)))
         body ...
         (sys-chdir cur)))))

  (if (eq? sys-symlink #f)
    (tar-condition-throw
     (<tar-not-support> "this system doesn't support symbolic link."
                        "(extract-tar-symboliclink name linkname)")))

  (with-current-directory
   (let ((new-name linkname))
     (receive (dir name-only ext)
         (decompose-path linkname)
       (if (string=? dir "/")
         (set! new-name (format #f "~a.~a" name-only ext)))
       (sys-chdir (sys-dirname name))
       (sys-symlink new-name (sys-basename name))))))

(define (extract-tar-hardlink name linkname)
  (if (eq? sys-symlink #f)
    (tar-condition-throw
     (<tar-not-support> "this system doesn't support hardlink."
                        "(extract-tar-hardlink name linkname)")))

  (if (char=? (string-ref linkname 0) #\/)
    (set! linkname (format #f "~a/~a" (sys-dirname name)
                           (sys-basename linkname))))
  (sys-link linkname name))

(define-method extract-tar-fifo ((tar <tar>) (entry <tar-entry>))
  (sys-mkfifo (~ entry 'name) #o666)
  (set! (~ tar 'skip-data?) #f))

(define-method extract-tar-data ((tar <tar>) (entry <tar-entry>) :key (output #f))
  (cond
   ((or (string=? #0=(~ entry 'typeflag) "0") (string=? #0# "\0")
        (string=? #0# "S"))
    (extract-tar-regular-file tar entry :output output))
   ((string=? #0# "1")
    (set! (~ tar 'post-processes)
          (append
           (list (list extract-tar-hardlink (~ entry 'name) (~ entry 'linkname)))
           (~ tar 'post-processes))))
   ((string=? #0# "2")
    (set! (~ tar 'post-processes)
          (append
           (list (list extract-tar-symboliclink (~ entry 'name) (~ entry 'linkname)))
           (~ tar 'post-processes))))
   ((or (string=? #0# "3") (string=? #0# "4"))
    (tar-condition-throw
     (<tar-not-support>
      "character/block file doesn't support due to lack of Gauche API.\n"
      "extract-tar-data ((tar <tar>) (entry <tar-entry>) post-process")))
   ((or (string=? #0# "5") (string=? #0# "D"))
    (make-tar-directory tar entry))
   ((string=? #0# "6")
    (extract-tar-fifo tar entry))
   (else
    (tar-condition-throw
     (<tar-not-support>
      #"Not supported yet. type:[~(~ entry 'typeflag)]\n"
      "extract-tar-data ((tar <tar>) (entry <tar-entry>) post-process")))))

(define-method extract ((entry <tar-entry>))
  (guard
      (exc
       ((tar-warning? exc)
        (format (current-error-port) "~a (~a)\n"
                (tar-warning-message exc) (tar-warning-debug-info exc))))
    (if (is-a? (~ entry 'tarobj) <tar-random-reader-base>)
      (port-seek (~ entry 'tarobj 'ioport)
                 (~ entry 'random-access-offset)
                 SEEK_SET))
    (extract-tar-data (~ entry 'tarobj) entry)))

(define-method extract ((entry <tar-entry>) (output <port>))
  (guard
      (exc
       ((tar-warning? exc)
        (format (current-error-port) "~a (~a)\n"
                (tar-warning-message exc) (tar-warning-debug-info exc))))
    (begin
      (if (not (output-port? output))
        (tar-condition-throw
         (<tar-error>
          "output port required."
          "extract-to-buffer ((entry <tar-entry>) (output <port>))")))
      (if (is-a? (~ entry 'tarobj) <tar-random-reader-base>)
        (port-seek (~ entry 'tarobj 'ioport)
                   (~ entry 'random-access-offset)
                   SEEK_SET))
      (extract-tar-data (~ entry 'tarobj) entry :output output))))

(define-method extract-generator ((entry <tar-entry>) :key (buffer-size 512))
  (guard
      (exc
       ((tar-warning? exc)
        (format (current-error-port) "~a (~a)\n"
                (tar-warning-message exc) (tar-warning-debug-info exc))))
    (let ((checked-buffer-size buffer-size))
      (cond
       ((not (= (modulo buffer-size *tar-block-size*) 0))
        (tar-condition-throw
         (<tar-error> "buffer-size must be 512 or multiple of 512 size."
                      "extract-generator ((entry <tar-entry>) buffer-size)")))
       ((< buffer-size 0)
        (tar-condition-throw
         (<tar-error> "buffer-size must be a positive value."
                      "extract-generator ((entry <tar-entry>) buffer-size)")))
       ((= buffer-size 0)
        (set! checked-buffer-size *tar-block-size*))
       ((> buffer-size (~ entry 'size))
        (set! checked-buffer-size *tar-block-size*)))
      (cond
       ((or (string=? #0=(~ entry 'typeflag) "0") (string=? #0# "\0"))
        (extract-tar-regular-file-generator
         (~ entry 'tarobj) entry checked-buffer-size))
       (else
        (tar-condition-throw
         (<tar-not-support>
          #"This type of entry isn't supported for extract generation. type:[~(~ entry 'typeflag)]\n"
          "extract-generator ((entry <tar-entry>) buffer-size)")))))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Write tar
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(define (new-pax-header-name path proc-number)
  (receive (dir name ext)
      (decompose-path path)
    ;; (format #t "~S ~S ~S ~S\n" proc-number dir name ext)
    (cond
     ((and (eq? name #f) (eq? ext #f))
      (format #f "./~a" (simplify-path
                         (format #f "./PaxHeaders.~d/~a" proc-number dir))))
     ((eq? ext #f)
      (simplify-path (format #f "~a/PaxHeaders.~d/~a" dir proc-number name)))
     ((eq? dir #f)
      (format #f "./PaxHeaders.~d/~a.~a" proc-number name ext))
     (else
      (if (string=? "." dir)
        (format #f "~a/PaxHeaders.~d/~a.~a" dir proc-number name ext) 
        (simplify-path
         (format #f "~a/PaxHeaders.~d/~a.~a" dir proc-number name ext)))))))

;; (format #t (new-pax-header-name "testdata/"))
;; (format #t (new-pax-header-name "testdata/fireworks.jpg"))
;; (format #t (new-pax-header-name "testdata/foobar"))
;; (format #t (new-pax-header-name "./testdata/"))
;; (format #t (new-pax-header-name "./testdata/fireworks.jpg"))
;; (format #t (new-pax-header-name "./testdata/foobar"))

(define-method set-tar-file-info ((tar <tar>) (header <tar-header-base>) src)
  (set! (~ header 'name) src)
  (set! (~ header 'mode) (logand (file-mode src) #o007777))
  (set! (~ header 'uid) (file-uid src))
  (set! (~ header 'gid) (file-gid src))
  (set! (~ header 'size) (if (> (file-size src) (expt 8 12))
                           (- (expt 8 12) 1)
                           (file-size src)))
  (set! (~ header 'mtime) (file-mtime src))
  (set! (~ header 'typeflag) "0")
  (set! (~ header 'uname) (sys-uid->user-name (file-uid src)))
  (set! (~ header 'gname) (sys-gid->group-name (file-gid src)))
  (set! (~ header 'chksum) (calc-chksum header)))

(define-method set-tar-file-info ((tar <tar>) (header <tar-header-base>) (ventry <tar-virtual-entry>))
  (set! (~ header 'name) (~ ventry 'name))
  (set! (~ header 'mode) (logand (~ ventry 'mode) #o007777))
  (set! (~ header 'uid) #0=(~ ventry 'uid))
  (set! (~ header 'gid) #1=(~ ventry 'gid))
  (set! (~ header 'size) (if (> (~ ventry 'size) (expt 8 12))
                           (- (expt 8 12) 1)
                           (~ ventry 'size)))
  (set! (~ header 'mtime) (~ ventry 'mtime))
  (set! (~ header 'typeflag) "0")
  (set! (~ header 'uname) (sys-uid->user-name #0#))
  (set! (~ header 'gname) (sys-gid->group-name #1#))
  (set! (~ header 'chksum) (calc-chksum header)))

(define-method set-tar-hardlink-info ((tar <tar>) (header <tar-header-base>) src dest)
  (set! (~ header 'name) src)
  (set! (~ header 'mode) (logand (file-mode src) #o007777))
  (set! (~ header 'uid) (file-uid src))
  (set! (~ header 'gid) (file-gid src))
  (set! (~ header 'size) 0)
  (set! (~ header 'mtime) (file-mtime src))
  (set! (~ header 'typeflag) "1")
  (set! (~ header 'uname) (sys-uid->user-name (file-uid src)))
  (set! (~ header 'gname) (sys-gid->group-name (file-gid src)))
  (set! (~ header 'linkname) dest)
  (set! (~ header 'chksum) (calc-chksum header)))

(define-method set-tar-symboliclink-info ((tar <tar>) (header <tar-header-base>) src)
  (set! (~ header 'name) src)
  (set! (~ header 'mode) (logand (file-mode src) #o007777))
  (set! (~ header 'uid) (file-uid src))
  (set! (~ header 'gid) (file-gid src))
  (set! (~ header 'size) 0)
  (set! (~ header 'mtime) (file-mtime src))
  (set! (~ header 'typeflag) "2")
  (set! (~ header 'uname) (sys-uid->user-name (file-uid src)))
  (set! (~ header 'gname) (sys-gid->group-name (file-gid src)))
  (set! (~ header 'linkname) (sys-readlink src))
  (set! (~ header 'chksum) (calc-chksum header)))

(define-method set-tar-characterfile-info ((tar <tar>) (header <tar-header-base>) src)
  (define (major src)
    (let ((stat (sys-stat src)))
      (ash (slot-ref stat 'rdev) -8)))
  (define (minor src)
    (let ((stat (sys-stat src)))
      (logand (slot-ref stat 'rdev) #xff)))
  
  (set! (~ header 'name) src)
  (set! (~ header 'mode) (logand (file-mode src) #o007777))
  (set! (~ header 'uid) (file-uid src))
  (set! (~ header 'gid) (file-gid src))
  (set! (~ header 'size) (file-size src))
  (set! (~ header 'mtime) (file-mtime src))
  (set! (~ header 'typeflag) "3")
  (set! (~ header 'uname) (sys-uid->user-name (file-uid src)))
  (set! (~ header 'gname) (sys-gid->group-name (file-gid src)))
  (set! (~ header 'devmajor) (major src))
  (set! (~ header 'devminor) (minor src))
  (set! (~ header 'chksum) (calc-chksum header)))

(define-method set-tar-blockfile-info ((tar <tar>) (header <tar-header-base>) src)
  (define (major src)
    (let ((stat (sys-stat src)))
      (ash (slot-ref stat 'rdev) -8)))
  (define (minor src)
    (let ((stat (sys-stat src)))
      (logand (slot-ref stat 'rdev) #xff)))
  
  (set! (~ header 'name) src)
  (set! (~ header 'mode) (logand (file-mode src) #o007777))
  (set! (~ header 'uid) (file-uid src))
  (set! (~ header 'gid) (file-gid src))
  (set! (~ header 'size) (file-size src))
  (set! (~ header 'mtime) (file-mtime src))
  (set! (~ header 'typeflag) "4")
  (set! (~ header 'uname) (sys-uid->user-name (file-uid src)))
  (set! (~ header 'gname) (sys-gid->group-name (file-gid src)))
  (set! (~ header 'devmajor) (major src))
  (set! (~ header 'devminor) (minor src))
  (set! (~ header 'chksum) (calc-chksum header)))

(define-method set-tar-directory-info ((tar <tar>) (header <tar-header-base>) src)
  (set! (~ header 'name) src)
  (set! (~ header 'mode) (logand (file-mode src) #o007777))
  (set! (~ header 'uid) (file-uid src))
  (set! (~ header 'gid) (file-gid src))
  (set! (~ header 'size) 0)
  (set! (~ header 'mtime) (file-mtime src))
  (set! (~ header 'typeflag) "5")
  (set! (~ header 'uname) (sys-uid->user-name (file-uid src)))
  (set! (~ header 'gname) (sys-gid->group-name (file-gid src)))
  (set! (~ header 'chksum) (calc-chksum header)))

(define-method set-tar-directory-info ((tar <tar>) (header <tar-header-base>) (ventry <tar-virtual-entry>))
  (set! (~ header 'name) (~ ventry 'name))
  (set! (~ header 'mode) (logand (~ ventry 'mode) #o007777))
  (set! (~ header 'uid) #0=(~ ventry 'uid))
  (set! (~ header 'gid) #1=(~ ventry 'gid))
  (set! (~ header 'size) (~ ventry 'size))
  (set! (~ header 'mtime) (~ ventry 'mtime))
  (set! (~ header 'typeflag) "5")
  (set! (~ header 'uname) (sys-uid->user-name #0#))
  (set! (~ header 'gname) (sys-gid->group-name #1#))
  (set! (~ header 'chksum) (calc-chksum header)))

(define-method set-tar-fifo-info ((tar <tar>) (header <tar-header-base>) src)
  (set! (~ header 'name) src)
  (set! (~ header 'mode) (logand (file-mode src) #o007777))
  (set! (~ header 'uid) (file-uid src))
  (set! (~ header 'gid) (file-gid src))
  (set! (~ header 'size) 0)
  (set! (~ header 'mtime) (file-mtime src))
  (set! (~ header 'typeflag) "6")
  (set! (~ header 'uname) (sys-uid->user-name (file-uid src)))
  (set! (~ header 'gname) (sys-gid->group-name (file-gid src)))
  (set! (~ header 'chksum) (calc-chksum header)))

(define-method write-pax-global-extend-header ((tar <tar>) key-value-alist :key (proc-number 12345))
  (let ((header (make <tar-header-posix>)) 
        (out (open-output-string))
        (string ""))

    (dolist (kv key-value-alist)
      (format out "~a" (make-pax-extend-header-record
                        (x->string key) (x->string value))))

    (set! string (get-output-string out))

    (set-tar-type header 'tar-posix)
    (set! (~ header 'name) (new-pax-header-name "global_header" proc-number))
    (set! (~ header 'mode) #o007777)
    (set! (~ header 'uid) 0)
    (set! (~ header 'gid) 0)
    (set! (~ header 'mtime) 0)
    (set! (~ header 'typeflag) "g")
    (set! (~ header 'uname) "0")
    (set! (~ header 'gname) "0")
    (set! (~ header 'linkname) "")
    (set! (~ header 'size) (string-length string))
    (set! (~ header 'chksum) (calc-chksum header))

    (write-tar-header tar header)
    (let* ((block-count 0)
           (buf (make-uvector <u8vector> *tar-block-size* 0)))
      (set! block-count (tar-block-count (string-length string)))
      (call-with-input-string string
        (^(in)
          (dotimes (i block-count)
            (uvector-copy! buf 0 (read-uvector <u8vector> *tar-block-size* in))
            (write-uvector buf (~ tar 'ioport) 0 *tar-block-size*)
            (uvector-copy! buf 0 *tar-eob*)))))))

(define-method write-pax-extend-header ((tar <tar>) (src <string>) :key (dest #f) (special-file? #f) (proc-number 12345))
  (define (major src)
    (let ((stat (sys-stat src)))
      (ash (slot-ref stat 'rdev) -8)))
  (define (minor src)
    (let ((stat (sys-stat src)))
      (logand (slot-ref stat 'rdev) #xff)))
  
  (let ((header (make <tar-header-posix>)) 
        (out (open-output-string))
        (string ""))

    (if (> (string-length src) 100)
      (format out "~a" (make-pax-extend-header-record "path" src)))
    (if (and (not (eq? dest #f)) (> (string-length dest) 100))
      (format out "~a" (make-pax-extend-header-record "linkpath" dest)))
    (if (> (file-size src) (expt 8 12))
      (format out "~a" (make-pax-extend-header-record
                        "size" (x->string (file-size src)))))
    (if (> (file-uid src) (expt 8 12))
      (format out "~a" (make-pax-extend-header-record
                        "uid" (x->string (file-uid src)))))
    (if (> (file-gid src) (expt 8 12))
      (format out "~a" (make-pax-extend-header-record
                        "gid" (x->string (file-gid src)))))
    (if (> (string-length (sys-uid->user-name (file-uid src))) 32)
      (format out "~a" (make-pax-extend-header-record
                        "uname" (sys-uid->user-name (file-uid src)))))
    (if (> (string-length (sys-gid->group-name (file-gid src))) 32)
      (format out "~a" (make-pax-extend-header-record
                        "gname" (sys-gid->group-name (file-gid src)))))

    (write-pax-times src out)
    (set! string (get-output-string out))

    (set-tar-type header 'tar-posix)
    (set! (~ header 'name) (new-pax-header-name src proc-number))
    (set! (~ header 'mode) (logand (file-mode src) #o007777))
    (set! (~ header 'uid) 0)
    (set! (~ header 'gid) 0)
    (set! (~ header 'mtime) (file-mtime src))
    (set! (~ header 'typeflag) "x")
    (set! (~ header 'uname) "0")
    (set! (~ header 'gname) "0")
    (set! (~ header 'linkname) (or dest ""))
    (set! (~ header 'size) (string-length string))

    (if (eq? special-file? #t)
      (set! (~ header 'devmajor) (major src))
      (set! (~ header 'devminor) (minor src)))

    (set! (~ header 'chksum) (calc-chksum header))

    (write-tar-header tar header)
    (let* ((block-count 0)
           (buf (make-uvector <u8vector> *tar-block-size* 0)))
      (set! block-count (tar-block-count (string-length string)))
      (call-with-input-string string
        (^(in)
          (dotimes (i block-count)
            (uvector-copy! buf 0 (read-uvector <u8vector> *tar-block-size* in))
            (write-uvector buf (~ tar 'ioport) 0 *tar-block-size*)
            (uvector-copy! buf 0 *tar-eob*)))))))

(define-method write-pax-extend-header ((tar <tar>) (ventry <tar-virtual-entry>) :key (dest #f) (special-file? #f) (proc-number 12345))
  (let ((header (make <tar-header-posix>)) 
        (out (open-output-string))
        (string ""))

    (if (> (string-length (~ ventry 'name)) 100)
      (format out "~a" (make-pax-extend-header-record "path" (~ ventry 'name))))
    ;; (if (and (not (eq? dest #f)) (> (string-length dest) 100))
    ;;   (format out "~a" (make-pax-extend-header-record "linkpath" dest)))
    (if (> (~ ventry 'size) (expt 8 12))
      (format out "~a" (make-pax-extend-header-record
                        "size" (x->string (~ ventry 'size)))))
    (if (> (~ ventry 'uid) (expt 8 12))
      (format out "~a" (make-pax-extend-header-record
                        "uid" (x->string (~ ventry 'uid)))))
    (if (> (~ ventry 'gid) (expt 8 12))
      (format out "~a" (make-pax-extend-header-record
                        "gid" (x->string (~ ventry 'gid)))))
    (if (> (string-length (sys-uid->user-name (~ ventry 'uid))) 32)
      (format out "~a" (make-pax-extend-header-record
                        "uname" (sys-uid->user-name (~ ventry 'uid)))))
    (if (> (string-length (sys-gid->group-name (~ ventry 'gid))) 32)
      (format out "~a" (make-pax-extend-header-record
                        "gname" (sys-gid->group-name (~ ventry 'gid)))))

    (write-pax-times ventry out)
    (set! string (get-output-string out))

    (set-tar-type header 'tar-posix)
    (set! (~ header 'name) (new-pax-header-name (~ ventry 'name) proc-number))
    (set! (~ header 'mode) (logand (~ ventry 'mode) #o007777))
    (set! (~ header 'uid) 0)
    (set! (~ header 'gid) 0)
    (set! (~ header 'mtime) (~ ventry 'mtime))
    (set! (~ header 'typeflag) "x")
    (set! (~ header 'uname) "0")
    (set! (~ header 'gname) "0")
    (set! (~ header 'linkname) (or dest ""))
    (set! (~ header 'size) (string-length string))

    (set! (~ header 'chksum) (calc-chksum header))

    (write-tar-header tar header)
    (let* ((block-count 0)
           (buf (make-uvector <u8vector> *tar-block-size* 0)))
      (set! block-count (tar-block-count (string-length string)))
      (call-with-input-string string
        (^(in)
          (dotimes (i block-count)
            (uvector-copy! buf 0 (read-uvector <u8vector> *tar-block-size* in))
            (write-uvector buf (~ tar 'ioport) 0 *tar-block-size*)
            (uvector-copy! buf 0 *tar-eob*)))))))

(define-method write-pax-times ((src <string>) out)
  (format out "~a" (make-pax-extend-header-record "mtime" (file-mtime src)
                                                  "~a=~a.000000000\n"))
  (format out "~a" (make-pax-extend-header-record "atime" (file-atime src)
                                                  "~a=~a.000000000\n"))
  (format out "~a" (make-pax-extend-header-record "ctime" (file-ctime src)
                                                  "~a=~a.000000000\n")))

(define-method write-pax-times ((ventry <tar-virtual-entry>) out)
  (format out "~a" (make-pax-extend-header-record "mtime" (~ ventry 'mtime)
                                                  "~a=~a.000000000\n"))
  (format out "~a" (make-pax-extend-header-record "atime" (~ ventry 'atime)
                                                  "~a=~a.000000000\n"))
  (format out "~a" (make-pax-extend-header-record "ctime" (~ ventry 'ctime)
                                                  "~a=~a.000000000\n")))

(define (make-pax-extend-header-record key value :optional (fmt "~a=~a\n"))
  (let* ((src (format #f fmt key value))
         (len (+ #0=(string-length src) 1
                 (+ (x->integer (floor (log #0# 10))) 1))) ;;; 1 meant space
         (new-src (format #f "~d ~a" len src)))
    (if (not (= (string-length new-src) len))
      (format #f "~d ~a" len src)
      new-src)))

(define-method write-tar-eob ((tar <tar>) (header <tar-header-base>))
  (clear-tar-header header)
  (dotimes (i 2)
    (write-uvector *tar-eob* (~ tar 'ioport))))

(define-method write-tar-header ((tar <tar>) (header <tar-header-base>))
  (write-uvector (~ header 'header) (~ tar 'ioport)))

(define-method write-tar-file-content ((tar <tar>) (header <tar-header-base>) src :optional (write-content? #t))
  (let ((block-count 0)
        (buf (make-uvector <u8vector> *tar-block-size* 0)))
    (set! block-count (tar-block-count (~ header 'size)))
    (write-tar-header tar header)
    (if write-content?
      (call-with-input-file src
        (^(in)
          (dotimes (i block-count)
            (uvector-copy! buf 0 (read-uvector <u8vector> *tar-block-size* in))
            (write-uvector buf (~ tar 'ioport) 0 *tar-block-size*)
            (uvector-copy! buf 0 *tar-eob*)))))))

(define-method write-tar-file-content ((tar <tar>) (header <tar-header-base>) (ventry <tar-virtual-entry>) :optional (write-content? #t))
  (write-tar-header tar header)
  (let* ((block-count (tar-block-count (~ ventry 'size)))
         (total-size (* *tar-block-size* block-count))
         (remain-size (- total-size (~ ventry 'size)))
         (remain-buf (make-uvector <u8vector> remain-size 0)))
    (if write-content?
      (begin
        (write-uvector (get-output-tar-virtual-entry ventry)
                          (~ tar 'ioport) 0 (~ ventry 'size))
        (write-uvector remain-buf (~ tar 'ioport) 0 remain-size)))))

(define-method write-tar-directory-like-content ((tar <tar>) (header <tar-header-base>) src)
  (write-tar-header tar header))

(define-method write-tar-content ((tar <tar>) (header <tar-header-posix>) (src <string>) ht)
  (let ((stat (sys-stat src))
        (proc-number (sys-getpid)))
    (clear-tar-header header)
    (set-tar-type header 'tar-posix)
    (cond
     ((and (eq? #0=(~ stat 'type) 'regular)
           (eq? (file-is-symlink? src) #f)
           (eq? (hash-table-exists? ht (~ stat 'ino)) #f))
      (write-pax-extend-header tar src :proc-number proc-number)
      (set-tar-file-info tar header src)
      (write-tar-file-content tar header src)
      (hash-table-push! ht (~ (sys-stat src) 'ino) src))
     ((and (eq? #0# 'regular) (eq? (file-is-symlink? src) #f)
           (eq? (hash-table-exists? ht (~ stat 'ino)) #t))
      (write-pax-extend-header tar src
                               :dest (car (hash-table-get ht (~ stat 'ino)))
                               :proc-number proc-number)
      (set-tar-hardlink-info tar header src
                             (car (hash-table-get ht (~ stat 'ino))))
      (write-tar-file-content tar header src #f)
      (hash-table-push! ht (~ (sys-stat src) 'ino) src))
     ((eq? (file-is-symlink? src) #t)
      (write-pax-extend-header tar src :dest (sys-readlink src)
                               :proc-number proc-number)
      (set-tar-symboliclink-info tar header src)
      (write-tar-directory-like-content tar header src))
     ((eq? #0# 'character)
      (write-pax-extend-header tar src :special-file? #t
                               :proc-number proc-number)
      (set-tar-characterfile-info tar header src)
      (write-tar-directory-like-content tar header src)
      (hash-table-push! ht (~ (sys-stat src) 'ino) src))
     ((eq? #0# 'block)
      (write-pax-extend-header tar src :special-file? #t
                               :proc-number proc-number)
      (set-tar-blockfile-info tar header src)
      (write-tar-directory-like-content tar header src)
      (hash-table-push! ht (~ (sys-stat src) 'ino) src))
     ((eq? #0# 'directory)
      (write-pax-extend-header tar src :proc-number proc-number)
      (set-tar-directory-info tar header src)
      (write-tar-directory-like-content tar header src)
      (hash-table-push! ht (~ (sys-stat src) 'ino) src))
     ((eq? #0# 'fifo)
      (write-pax-extend-header tar src :proc-number proc-number)
      (set-tar-fifo-info tar header src)
      (write-tar-directory-like-content tar header src))
     (else (tar-condition-throw
            (<tar-error>
             #"unknown type: ~src"
             "write-tar-content ((tar <tar>) src ht)"))))))

(define-method write-tar-content ((tar <tar>) (header <tar-header-posix>) (ventry <tar-virtual-entry>) ht)
  (let ((proc-number (sys-getpid)))
    (clear-tar-header header)
    (set-tar-type header 'tar-posix)
    (cond
     ((eq? (~ ventry 'type) 'regular)
      (write-pax-extend-header tar ventry :proc-number proc-number)
      (set-tar-file-info tar header ventry)
      (write-tar-file-content tar header ventry))
     ((eq? (~ ventry 'type) 'directory)
      (write-pax-extend-header tar ventry :proc-number proc-number)
      (set-tar-directory-info tar header ventry)
      (write-tar-directory-like-content tar header ventry))
     (else (tar-condition-throw
            (<tar-error>
             #"unknown type: ~src"
             "write-tar-content ((tar <tar>) src ht)"))))))

(define (open-output-tar-writer src)
  (let ((tar (make <tar-writer>)))
    (set! (~ tar 'name) src)
    tar))

(define-method close-output-tar-writer ((tar <tar-writer>))
  (set! (~ tar 'ioport) (open-output-file (~ tar 'name)))
  (unwind-protect
   (let ((tar-header (make <tar-header-posix>)))
     (dolist (file (reverse (~ tar 'filelist)))
       (write-tar-content tar tar-header file (~ tar 'hardlink-table)))
     (write-tar-eob tar tar-header))
   (close-output-port (~ tar 'ioport))))

(define-method add-content ((tar <tar-writer-base>) src)
  (set! #0=(~ tar 'filelist) (append (list src) #0#)))

(define (open-output-tar-builder src)
  (let ((tar (make <tar-builder>)))
    (set! (~ tar 'name) src)
    (set! (~ tar 'ioport) (open-output-file (~ tar 'name)))
    tar))

(define-method close-output-tar-builder ((tar <tar-builder>))
  (unwind-protect
   (let ((tar-header (make <tar-header-posix>)))
     (write-tar-eob tar tar-header))
   (close-output-port (~ tar 'ioport))))

(define-method add-content ((tar <tar-builder>) src)
  (let ((tar-header (make <tar-header-posix>)))
    (write-tar-content tar tar-header src (~ tar 'hardlink-table))))

(define (call-with-output-tar-writer src proc)
  (let ((tar (open-output-tar-writer src)))
    (unwind-protect
     (proc tar)
     (close-output-tar-writer tar))))

(define (call-with-output-tar-builder src proc)
  (let ((tar (open-output-tar-builder src)))
    (unwind-protect
     (proc tar)
     (close-output-tar-builder tar))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Read tar-gz
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(define (open-input-tar-gz src :key (debug #f))
  (let ((tar-gz (if debug (make <debug-tar-gz-reader>) (make <tar-gz-reader>))))
    (set! (~ tar-gz 'name) src)
    (set! (~ tar-gz 'gzip-dorain-port) (open-input-file src))
    (set! (~ tar-gz 'ioport) (open-inflating-port (~ tar-gz 'gzip-dorain-port)
                                                  :window-bits 31))
    tar-gz))

(define (open-input-tar-gz-from-port port :key (debug #f))
  (if (not (eq? (input-port? port) #t))
    (tar-condition-throw
     (<tar-error>
      "argument `port` must be input port."
      "open-input-tar-gz-from-port port :key (debug #f)")))
  (let ((tar-gz (if debug (make <debug-tar-gz-reader>) (make <tar-gz-reader>))))
    (set! (~ tar-gz 'name) (port-name port))
    (set! (~ tar-gz 'gzip-dorain-port) port)
    (set! (~ tar-gz 'ioport) (open-inflating-port (~ tar-gz 'gzip-dorain-port)
                                                  :window-bits 31))
    tar-gz))

(define-method close-input-tar-gz ((tar <tar-gz-reader-base>))
  (close-input-port (~ tar 'gzip-dorain-port))
  (set! (~ tar 'name) "")
  (if (not (eq? (~ tar 'post-processes) '()))
    (dolist (f (~ tar 'post-processes))
      (apply (car f) (cdr f)))))

(define (call-with-input-tar-gz src proc :key (debug #f))
  (let ((tar-gz (open-input-tar-gz src :debug debug)))
    (unwind-protect
     (proc tar-gz)
     (close-input-tar-gz tar-gz))))

(define (call-with-input-tar-gz-from-port port proc :key (debug #f))
  (let ((tar-gz (open-input-tar-gz-from-port port :debug debug)))
    (unwind-protect
     (proc tar-gz)
     (close-input-tar-gz tar-gz))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Write tar-gz
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(define (open-output-tar-gz-writer src :key (compression-level 6))
  (let ((tar-gz (make <tar-gz-writer>)))
    (set! (~ tar-gz 'name) src)
    (set! (~ tar-gz 'compression-level) compression-level)
    tar-gz))

(define-method close-output-tar-gz-writer ((tar-gz <tar-gz-writer>))
  (set! (~ tar-gz 'gzip-dorain-port) (open-output-file (~ tar-gz 'name)))
  (set! (~ tar-gz 'ioport) (open-deflating-port
                            (~ tar-gz 'gzip-dorain-port)
                            :window-bits 31
                            :compression-level (~ tar-gz 'compression-level)))
  (unwind-protect
   (let ((tar-header (make <tar-header-posix>)))
     (dolist (file (reverse (~ tar-gz 'filelist)))
       (write-tar-content tar-gz tar-header file (~ tar-gz 'hardlink-table)))
     (write-tar-eob tar-gz tar-header))
   (deflating-port-full-flush (~ tar-gz 'ioport))
   (close-output-port (~ tar-gz 'ioport))
   (close-output-port (~ tar-gz 'gzip-dorain-port))))

(define (open-output-tar-gz-builder src :key (compression-level 6))
  (let ((tar-gz (make <tar-gz-builder>)))
    (set! (~ tar-gz 'name) src)
    (set! (~ tar-gz 'gzip-dorain-port) (open-output-file src))
    (set! (~ tar-gz 'compression-level) compression-level)
    (set! (~ tar-gz 'ioport)
          (open-deflating-port
           (~ tar-gz 'gzip-dorain-port)
           :window-bits 31
           :compression-level (~ tar-gz 'compression-level)))
    tar-gz))

(define-method close-output-tar-gz-builder ((tar-gz <tar-gz-builder>))
  (unwind-protect
   (let ((tar-header (make <tar-header-posix>)))
     (write-tar-eob tar-gz tar-header))
   (deflating-port-full-flush (~ tar-gz 'ioport))
   (close-output-port (~ tar-gz 'ioport))
   (close-output-port (~ tar-gz 'gzip-dorain-port))))

(define-method add-content ((tar-gz <tar-gz-builder>) src)
  (let ((tar-header (make <tar-header-posix>)))
    (write-tar-content tar-gz tar-header src (~ tar-gz 'hardlink-table))))

(define (call-with-output-tar-gz-writer src proc :key (compression-level 6))
  (let ((tar-gz (open-output-tar-gz-writer
                 src :compression-level compression-level)))
    (unwind-protect
     (proc tar-gz)
     (close-output-tar-gz-writer tar-gz))))

(define (call-with-output-tar-gz-builder src proc :key (compression-level 6))
  (let ((tar-gz (open-output-tar-gz-builder
                 src :compression-level compression-level)))
    (unwind-protect
     (proc tar-gz)
     (close-output-tar-gz-builder tar-gz))))

