;;;
;;; Test tar
;;;

(use gauche.test)
(use file.util)
(use rfc.md5)
(use util.digest)
(use gauche.collection)

(define *md5* '(("testdata/archive.c" . "4ae32db053101e3e2138504de9a776e6")
                ("testdata/archive_class.c" . "c9d6a1504421a976e8cedeb8e093f976")
                ("testdata/archive_proc.c" . "3d0ad9c0acf038c229b10e17519f6663")
                ("testdata/fireworks.jpg" . "5b23602a16aefbf9e482f63857b52fa8")))

(define *sparse-md5* '(("dumptar.c" . "91539ec6141869a9ea9ffe0b6c663fd6")
                       ("sparsefile" . "1afee6eebf172a3188c7dfafdfdbbba2")))

(define (calc-md5-from-file src)
  (calc-md5-from-string (file->string src)))

(define (calc-md5-from-string src)
  (digest-hexify (md5-digest-string src)))

(define-syntax with-current-directory
    (syntax-rules ()
      ((_ body ...)
       (let ((cur (current-directory)))
         body ...
         (sys-chdir cur)))))

(test-start "tar")
(use tar)
(test-module 'tar)

;; The following is a dummy test code.
;; Replace it for your tests.
;; (test* "test-tar" "tar is working"
;;        (test-tar))

(dolist (tar-file '("./testdata/ztestdatav7.tar" "./testdata/ztestdataustar.tar"
                    "./testdata/ztestdatagnuold.tar"
                    "./testdata/ztestdatagnu.tar"
                    "./testdata/ztestdataposix.tar"))
  (call-with-input-tar
   tar-file
   (^(tar)
     (map (^(entry)
            (test* #"tar fast skip ~tar-file ~(~ entry 'name)" #t #t))
          tar))
   :fast-skip? #t))

(dolist (tar-file '("./testdata/ztestdatav7.tar" "./testdata/ztestdataustar.tar"
                    "./testdata/ztestdatagnuold.tar"
                    "./testdata/ztestdatagnu.tar"
                    "./testdata/ztestdataposix.tar"))
  (call-with-input-tar
   tar-file
   (^(tar)
     (map (^(entry)
            (let ((md5 (assoc (~ entry 'name) *md5*))
                  (out #f))
              (if md5
                (begin
                  (set! out (open-output-string))
                  (extract entry out)
                  (test* #"extract tar content created by GNU Tar. file: ~tar-file @ ~(car md5)"
                         #t
                         (string=? (cdr md5) (calc-md5-from-string (get-output-string out)))))
                (test* #"extract tar content created by GNU Tar. file: ~tar-file @ testdata/"
                       #t
                       (string=? (~ entry 'name) "testdata/")))))
          tar))))

(define *long-dir* "./aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa")
(dolist (tar-file '("zlonggnuold.tar" "zlonggnu.tar" "zlongposix.tar"))
  (with-current-directory
   (sys-chdir "testdata")
   (call-with-input-tar tar-file (^(tar) (map extract tar)))
   (unwind-protect
    (begin
      (dolist (path (list (format #f "~a/src.txt" *long-dir*)
                          (format #f "~a/dest.txt" *long-dir*)
                          (format #f "~a/long.txt" *long-dir*)
                          (format #f "~a/longlink.txt" *long-dir*)))
        (test* #"long directory check ~tar-file @ ~path" #t (file-exists? path)))
      (test* "symbolic link?" #t
             (file-is-symlink? (format #f "~a/longlink.txt" *long-dir*)))))
   (delete-directory* *long-dir*)))

(dolist (tar-file '("./testdata/zsparseoldgnu.tar" "./testdata/zsparsegnu.tar"
                    "./testdata/zsparseposix00.tar" "./testdata/zsparseposix01.tar"
                    "./testdata/zsparseposix10.tar"))
  (call-with-input-tar
   tar-file
   (^(tar)
     (map (^(entry)
            (let ((md5 (assoc (~ entry 'name) *sparse-md5*))
                  (out #f))
              (if md5
                (begin
                  (set! out (open-output-string))
                  (extract entry out)
                  (test* #"extract sparse file created by GNU Tar. file: ~tar-file @ ~(car md5)"
                         #t
                         (string=?
                          (cdr md5)
                          (calc-md5-from-string (get-output-string out))))))))
          tar))))

(test* "tar virtual entry test" "3858f62230ac3c915f300c664312c63f"
       (let ((out (open-output-tar-virtual-entry "foo.txt"))
              (empty-dir (make-tar-virtual-entry-directory "temp/"))
              (tar (open-output-tar-gz-writer "ventry.tar.gz"))
              (md5 ""))
          (format out "foobar")
          (add-content tar out)
          (add-content tar empty-dir)
          (close-output-tar-gz-writer tar)
          (close-output-tar-virtual-entry out)
          (call-with-input-tar-gz "ventry.tar.gz"
                                  (^(tgz)
                                    (map (^(entry)
                                          (if (string=? (~ entry 'name) "foo.txt")
                                            (let ((out (open-output-string)))
                                              (extract entry out)
                                              (set! md5 (calc-md5-from-string (get-output-string out))))))
                                         tgz)))
          (delete-file "ventry.tar.gz")
          md5))

;; If you don't want `gosh' to exit with nonzero status even if
;; the test fails, pass #f to :exit-on-failure.
(test-end :exit-on-failure #t)




